#!/usr/bin/env python3
# Copyright 2020 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import argparse
import logging
import os
import sys
import yaml
from .config_generator import ConfigProcessor
from multiprocessing import Pool, cpu_count

logger = logging.getLogger(__name__)


class Loader(yaml.SafeLoader):
    """
    Overloading the default YAML Loader by adding the custom include tag
    """

    ROOT_DIR = os.path.abspath(os.curdir)

    def __init__(self, stream):

        self._root = os.path.split(stream.name)[0]
        super(Loader, self).__init__(stream)

    def include(self, node):
        """
        Method implementing the custom include tag logic that will grab a value from a yaml key in a specified folder
        at a specified path
        :param node: String containing the path and key variables
        :return: String values representing the extracted value for the specified path, key combination under node
        """
        path, key = self.construct_yaml_str(node).split(" ")
        filename = os.path.join(self.ROOT_DIR, path)

        with open(filename, 'r') as f:
            yaml_structure = yaml.load(f, Loader)
            return self.__traverse_path(path=key, yaml_dict=yaml_structure)

    def __traverse_path(self, path: str, yaml_dict: dict):
        """
        Method for safe traversing a yaml dictionary to extract a value from a key path
        :param path: String representing the keys needed to traverse the dictionary
        :param yaml_dict: YAML dictionary
        :return: String representing the value for the key path specified
        """
        keys = path.split(".")

        current_key = keys.pop(0)

        if current_key in yaml_dict:
            if len(keys) == 0:
                return yaml_dict[current_key]
            else:
                if isinstance(yaml_dict[current_key], dict):
                    return self.__traverse_path(path=".".join(keys), yaml_dict=yaml_dict[current_key])
                else:
                    raise Exception("{1}[{0}] is not traversable.".format(
                        current_key, yaml_dict))
        else:
            raise Exception("Key not found for {0} in dictionary {1}.".format(
                current_key, yaml_dict))


def merge_configs(directories, levels, output_dir, enable_parallel):
    """
    Method for running the merge configuration logic under different formats
    :param directories: list of paths for leaf directories
    :param levels: list of hierarchy levels to traverse
    :param output_dir: where to save the generated configs
    :param enable_parallel: to enable parallel config generation
    """
    config_processor = ConfigProcessor()
    process_config = []
    for path in directories:
        process_config.append((config_processor, path, levels, output_dir))

    if enable_parallel:
        logger.info("Processing config in parallel")
        with Pool(cpu_count()) as p:
            p.map(merge_logic, process_config)
    else:
        for config in process_config:
            merge_logic(config)


def merge_logic(process_params):
    """
    Method implementing the merge config logic
    :param process_params: tuple that contains config for running the merge_logic
    """
    config_processor = process_params[0]
    path = process_params[1]
    levels = process_params[2]
    output_dir = process_params[3]

    # load the !include tag
    Loader.add_constructor('!include', Loader.include)

    # override the Yaml SafeLoader with our custom loader
    yaml.SafeLoader = Loader

    # for path in directories:
    # use the HIML deep merge functionality
    output = dict(
        config_processor.process(path=path, output_format="yaml", print_data=False, multi_line_string=True))
    # exchange the levels to which to run for with the values extracted from the yaml structure
    level_values = [output.get(level) for level in levels]

    # create the publish path and all level_values except the last one
    publish_path = os.path.join(output_dir, '') + '/'.join(level_values[:-1])
    if not os.path.exists(publish_path):
        os.makedirs(publish_path)

    # create the yaml file for output using the publish_path and last level_values element
    filename = "{0}/{1}.yaml".format(publish_path, level_values[-1])
    logger.info("Found input config directory: %s", path)
    logger.info("Storing generated config to: %s", filename)
    with open(filename, "w+") as f:
        f.write(yaml.dump(output))


def is_leaf_directory(dir, leaf_directories):
    return any(dir.startswith(leaf) for leaf in leaf_directories)


def get_leaf_directories(src, leaf_directories):
    """
    Method for doing a deep search of directories matching either the desired
    leaf directories.
    :param src: the source path to start looking from
    :return: the list of absolute paths
    """
    directories = []

    for root, dirs, files in os.walk(src):
        for dr in dirs:
            # if directory is hidden skip
            if dr.startswith('.') or root.split("/")[1].startswith("."):
                continue
            elif is_leaf_directory(dr, leaf_directories):
                directory = root + "/" + dr
                directories.append(directory)
            else:
                continue

    if len(directories) == 0:
        sys.exit("No leaf directories found")

    return directories


def parser_options(args):
    parser = argparse.ArgumentParser()
    parser.add_argument('path', type=str, help='The configs directory')

    parser.add_argument('--output-dir', dest='output_dir', type=str,
                        help='output directory, where generated configs will be saved', required=True)
    parser.add_argument('--levels', dest='hierarchy_levels', nargs='+',
                        help='hierarchy levels, for instance: env, region, cluster', required=True)
    parser.add_argument('--leaf-directories', dest='leaf_directories', nargs='+',
                        help='leaf directories, for instance: cluster', required=True)
    parser.add_argument('--enable-parallel', dest='enable_parallel', default=False,
                        action='store_true', help='Process config using multiprocessing')
    return parser.parse_args(args)


def run(args=None):
    opts = parser_options(args)

    # extract the list of absolute paths for leaf directories
    dirs = get_leaf_directories(opts.path, opts.leaf_directories)

    # merge the configs using HIML
    merge_configs(dirs, opts.hierarchy_levels,
                  opts.output_dir, opts.enable_parallel)



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import re

from .inject_env import EnvVarInjector
from .inject_secrets import SecretInjector
from .python_compat import iteritems, string_types, primitive_types


def is_interpolation(value):
    return isinstance(value, string_types) and '{{' in value and '}}' in value \
           and not is_escaped_interpolation(value)


def is_escaped_interpolation(value):
    return isinstance(value, string_types) and '{{`' in value and '`}}' in value


def is_fully_escaped_interpolation(value):
    return isinstance(value, string_types) and value.startswith('{{`') and value.endswith('`}}')


def is_full_interpolation(value):
    return is_interpolation(value) and value.startswith('{{') and value.endswith('}}') \
           and not is_fully_escaped_interpolation(value)


def remove_white_spaces(value):
    return re.sub(r"\s+", "", value)


def replace_parent_working_directory(value, cwd):
    if "{{cwd}}" in value:
        return value.replace("{{cwd}}", cwd)
    return value


class InterpolationResolver(object):

    def resolve_interpolations(self, data):
        # Resolve from dictionary. Do one iteration before secret resolving, in order to resolve interpolations such as
        # the aws.profile
        # Example:
        # my_profile: test
        # aws:
        #   profile: "{{my_profile}}"
        from_dict_injector = DictInterpolationResolver(data, FromDictInjector())
        from_dict_injector.resolve_interpolations(data)

        return data


class EscapingResolver(object):

    def resolve_escaping(self, data):
        """
        Should do one last check through all values to ensure the ones that were escaped are cleaned of the escape
        sequence
        """
        from_dict_injector = DictEscapingResolver(data, FromDictInjector())
        from_dict_injector.clean_escapes(data)

        return data


class SecretResolver(object):

    def resolve_secrets(self, data, default_aws_profile):
        # Resolve interpolations representing secrets
        # Example:
        # value1: "{{ssm.path(mysecret)}}"
        injector = SecretInjector(default_aws_profile)
        secrets_resolver = SecretsInterpolationResolver(injector)
        secrets_resolver.resolve_interpolations(data)

        return data


class EnvVarResolver(object):
    def resolve_env_vars(self, data):
        injector = EnvVarInjector()
        env_resolver = EnvVarInterpolationsResolver(injector)
        env_resolver.resolve_interpolations(data)
        return data


class DictIterator(object):

    def loop_all_items(self, data, process_func):
        if isinstance(data, string_types):
            return process_func(data)
        if isinstance(data, list):
            items = []
            for item in data:
                if isinstance(item, list):
                    items.extend(item)
                else:
                    items.append(self.loop_all_items(item, process_func))
            return items
        if isinstance(data, dict):
            for key in data:
                value = data[key]
                resolved_value = self.loop_all_items(value, process_func)
                data[key] = resolved_value
        return data


class AbstractEscapingResolver(DictIterator):

    def clean_escapes(self, data):
        return self.loop_all_items(data, self.clean_escape)

    def clean_escape(self, line):
        # Check if line is escaped
        if not is_escaped_interpolation(line):
            return line
        return self.do_clean_escapes(line)

    def do_clean_escapes(self, line):
        pass


class AbstractInterpolationResolver(DictIterator):

    def resolve_interpolations(self, data):
        return self.loop_all_items(data, self.resolve_interpolation)

    def resolve_interpolation(self, line):
        if not is_interpolation(line):
            return line
        if is_full_interpolation(line):
            line = remove_white_spaces(line)
        return self.do_resolve_interpolation(line)

    def do_resolve_interpolation(self, line):
        pass


class DictEscapingResolver(AbstractEscapingResolver):
    def __init__(self, data, from_dict_injector):
        AbstractEscapingResolver.__init__(self)
        self.data = data
        self.from_dict_injector = from_dict_injector
        self.full_blob_injector = FullBlobInjector()

    def do_clean_escapes(self, line):
        updated_line = self.from_dict_injector.resolve(line, self.data)
        return self.full_blob_injector.clean(updated_line)


class DictInterpolationResolver(AbstractInterpolationResolver):
    def __init__(self, data, from_dict_injector):
        AbstractInterpolationResolver.__init__(self)
        self.data = data
        self.from_dict_injector = from_dict_injector
        self.full_blob_injector = FullBlobInjector()

    def do_resolve_interpolation(self, line):
        updated_line = self.from_dict_injector.resolve(line, self.data)
        return self.full_blob_injector.resolve(updated_line, self.data)


class SecretsInterpolationResolver(AbstractInterpolationResolver):
    def __init__(self, secrets_injector):
        AbstractInterpolationResolver.__init__(self)
        self.secrets_injector = secrets_injector

    def do_resolve_interpolation(self, line):
        return self.secrets_injector.inject_secret(line)


class EnvVarInterpolationsResolver(AbstractInterpolationResolver):
    def __init__(self, env_vars_injector):
        AbstractInterpolationResolver.__init__(self)
        self.env_vars_injector = env_vars_injector

    def do_resolve_interpolation(self, line):
        return self.env_vars_injector.inject_env_var(line)


class InterpolationValidator(DictIterator):

    def check_all_interpolations_resolved(self, data):
        return self.loop_all_items(data, self.validate_value)

    def validate_value(self, value):
        if is_interpolation(value):
            raise Exception("Interpolation could not be resolved {} and strict validation was enabled.".format(value))
        return value


class FromDictInjector(object):

    def __init__(self):
        self.results = {}

    def resolve(self, line, data):
        """
        :param line: {{env.name}}
        :param data: (env: name: dev)
        :return: dev
        """

        self.parse_leaves(data, "")
        for key, value in iteritems(self.results):
            placeholder = "{{" + key + "}}"
            if placeholder not in line:
                continue
            elif isinstance(value, (int, bool)):
                return value
            elif not is_interpolation(value):
                line = line.replace(placeholder, value)
        return line

    def parse_leaves(self, data, partial_key):
        if isinstance(data, primitive_types):
            self.results[partial_key] = data
            return
        if isinstance(data, dict):
            for key in data:
                value = data[key]
                new_key = partial_key
                if new_key:
                    new_key += "."
                new_key += key
                self.parse_leaves(value, new_key)


class FullBlobInjector(object):

    def resolve(self, line, data):
        if not is_full_interpolation(line):
            return line

        keys = self.get_keys_from_interpolation(line)
        resolved_value = self.get_inner_value(keys, data)
        is_valid_value = resolved_value is not None and not is_interpolation(resolved_value)

        return resolved_value if is_valid_value else line

    def clean(self, line):
        # {{` something {{ value }} `}}
        if is_fully_escaped_interpolation(line):
            resolved_value = self.get_value_from_escaping(line)
            is_valid_value = resolved_value is not None

            return resolved_value if is_valid_value else line

        # before {{` {{value}} `}} after
        elif is_escaped_interpolation(line):
            prefix = line[0:line.find('{{`')]
            escaping_string = line[line.find('{{`'):line.find('`}}') + 3]
            suffix = line[line.find('`}}') + 3:len(line)]
            resolved_value = prefix + self.get_value_from_escaping(escaping_string) + suffix
            is_valid_value = resolved_value is not None

            return resolved_value if is_valid_value else line

        # nothing to clean
        return line

    @staticmethod
    def get_inner_value(keys, data):
        for key in keys:
            if key in data:
                data = data[key]
            else:
                return None

        return data

    @staticmethod
    def get_keys_from_interpolation(line):
        # remove {{ and }}
        line = line[2:-2]

        return line.split('.')

    @staticmethod
    def get_value_from_escaping(line):
        # remove {{` and `}}
        line = line[3:-3]

        return line




# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import argparse
import os
from .config_generator import ConfigProcessor
from enum import Enum

class ListMergeStrategy(Enum):
    append = 'append'
    override = 'override'
    prepend = 'prepend'
    append_unique = 'append_unique' #WARNING: currently this strategy does not support list of dicts, only list of str

    def __str__(self):
        return self.value

class ConfigRunner(object):

    def run(self, args):
        parser = self.get_parser()
        opts = parser.parse_args(args)
        self.do_run(opts)

    def do_run(self, opts):
        cwd = opts.cwd if opts.cwd else os.getcwd()
        filters = opts.filter if opts.filter else ()
        excluded_keys = opts.exclude if opts.exclude else ()
        if opts.output_file is None:
            opts.print_data = True

        config_processor = ConfigProcessor()
                                 
        config_processor.process(cwd, opts.path, filters, excluded_keys, opts.enclosing_key, opts.remove_enclosing_key,
                                 opts.output_format, opts.print_data, opts.output_file, opts.skip_interpolation_resolving,
                                 opts.skip_interpolation_validation, opts.skip_secrets, opts.multi_line_string,
                                 type_strategies= [(list, [opts.merge_list_strategy.value]), (dict, ["merge"])] )

    @staticmethod
    def get_parser(parser=None):
        if not parser:
            parser = argparse.ArgumentParser()
            parser.add_argument('path', type=str, help='The config directory')

        parser.add_argument('--output-file', dest='output_file', type=str,
                            help='output file location')
        parser.add_argument('--print-data', action='store_true',
                            help='print generated data on screen')
        parser.add_argument('--format', dest='output_format', type=str, default="yaml",
                            help='output file format')
        parser.add_argument('--filter', dest='filter', action='append',
                            help='keep these keys from the generated data')
        parser.add_argument('--exclude', dest='exclude', action='append',
                            help='exclude these keys from generated data')
        parser.add_argument('--skip-interpolation-validation', action='store_true',
                            help='will not throw an error if interpolations can not be resolved')
        parser.add_argument('--skip-secrets', action='store_true',
                            help='will not throw an error if secrets can not be resolved')
        parser.add_argument('--skip-interpolation-resolving', action='store_true',
                            help='do not perform any AWS calls to resolve interpolations')
        parser.add_argument('--enclosing-key', dest='enclosing_key', type=str,
                            help='enclose the generated data under a common key')
        parser.add_argument('--remove-enclosing-key', dest='remove_enclosing_key', type=str,
                            help='remove enclosed data from under a common key')
        parser.add_argument('--cwd', dest='cwd', type=str, default="",
                            help='the working directory')
        parser.add_argument('--multi-line-string', action='store_true',
                            help='will overwrite the global yaml dumper to use block style')
        parser.add_argument('--list-merge-strategy', dest='merge_list_strategy', type=ListMergeStrategy, choices=list(ListMergeStrategy),
                            default='append',
                            help='override default merge strategy for list')
        parser.add_argument('--version', action='version', version='%(prog)s v{version}'.format(version="0.15.0"),
                            help='print himl version')
        return parser

def run(args=None):
    ConfigRunner().run(args)



# Copyright 2021 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

from os import getenv


class EnvVarInjector(object):
    """
    Resolve variables in the form:
    {{env(HOME)}}
    """

    def __init__(self):
        return

    def is_interpolation(self, value):
        return value.startswith('{{') and value.endswith('}}')

    def inject_env_var(self, line):
        """
        Check if value is an interpolation and try to resolve it.
        """
        if not self.is_interpolation(line):
            return line

        # remove {{ and }}
        updated_line = line[2:-2]

        # check supported function to ensure the proper format is used
        if not self.is_env_interpolation(updated_line):
            return line

        # remove env( and ) to extract the env Variable
        updated_line = updated_line[4:-1]

        # If env variable is missing or not set, the output will be None
        return getenv(updated_line)

    def is_env_interpolation(self, value):
        return value.startswith('env(') and value.endswith(')')



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import json
import os
from collections import OrderedDict
import logging

import pathlib2
import yaml
from deepmerge import Merger

from .interpolation import InterpolationResolver, EscapingResolver, InterpolationValidator, SecretResolver, \
    DictIterator, replace_parent_working_directory, EnvVarResolver
from .python_compat import iteritems, primitive_types, PY3
from .remote_state import S3TerraformRemoteStateRetriever

logger = logging.getLogger(__name__)


class ConfigProcessor(object):

    def process(self, cwd=None,
                path=None,
                filters=(),
                exclude_keys=(),
                enclosing_key=None,
                remove_enclosing_key=None,
                output_format="yaml",
                print_data=False,
                output_file=None,
                skip_interpolations=False,
                skip_interpolation_validation=False,
                skip_secrets=False,
                multi_line_string=False,
                type_strategies=[(list, ["append_unique"]), (dict, ["merge"])],
                fallback_strategies=["override"],
                type_conflict_strategies=["override"]):

        path = self.get_relative_path(path)

        if skip_interpolations or skip_secrets:
            skip_interpolation_validation = True

        if cwd is None:
            cwd = os.getcwd()

        generator = ConfigGenerator(cwd, path, multi_line_string, type_strategies, fallback_strategies,
                                    type_conflict_strategies)
        generator.generate_hierarchy()
        generator.process_hierarchy()

        # Exclude data before interpolations
        if len(exclude_keys) > 0:
            generator.exclude_keys(exclude_keys)

        # Resolve multiple levels of interpolations:
        if not skip_interpolations:
            # TODO: reduce the number of calls to resolve_interpolations
            generator.resolve_interpolations()

            # Resolve nested interpolations:
            # Example:
            # map1:
            #    key1: value1
            # map2: "{{map1.key1}}"
            # value: "something-{{map2.key1}} <--- this will be resolved at this step
            generator.resolve_interpolations()

            # Add dynamic data and resolve interpolations using dynamic data:
            generator.add_dynamic_data()
            generator.resolve_interpolations()

            # Add env vars and resolve interpolations using env vars:
            generator.resolve_env()
            generator.resolve_interpolations()

            # Add secrets and resolve interpolations using secrets:
            if not skip_secrets:
                default_aws_profile = self.get_default_aws_profile(generator.generated_data)
                generator.resolve_secrets(default_aws_profile)
                # Perform resolving in case some secrets are used in nested interpolations.
                # Example:
                # value1: "{{ssm.mysecret}}"
                # value2: "something-{{value1}} <--- this will be resolved at this step
                generator.resolve_interpolations()

        # Filter data before interpolation validation
        if len(filters) > 0:
            generator.filter_data(filters)
        if not skip_interpolation_validation:
            generator.validate_interpolations()

        if enclosing_key:
            logger.info("Adding enclosing key {}".format(enclosing_key))
            data = generator.add_enclosing_key(enclosing_key)
        elif remove_enclosing_key:
            logger.info("Removing enclosing key {}".format(remove_enclosing_key))
            data = generator.remove_enclosing_key(remove_enclosing_key)
        else:
            data = generator.generated_data

        generator.clean_escape_characters()

        if print_data or output_file:
            formatted_data = generator.output_data(data, output_format)

            if print_data:
                print(formatted_data)

            if output_file:
                with open(output_file, 'w') as f:
                    f.write(formatted_data)

        return data

    @staticmethod
    def get_default_aws_profile(data):
        return data['aws']['profile'] if 'aws' in data and 'profile' in data['aws'] else None

    @staticmethod
    def get_relative_path(path):
        cwd = os.path.join(os.getcwd(), '')
        if path.startswith(cwd):
            return path[len(cwd):]
        return path


class ConfigGenerator(object):
    """
    this class is used to create a config generator object which will be used to generate cluster definition files
    from the hierarchy of directories. The class implements methods that performs deep merging on dicts so the end result
    will contain merged data on each layer.
    """

    def __init__(self, cwd, path, multi_line_string, type_strategies, fallback_strategies, type_conflict_strategies):
        self.cwd = cwd
        self.path = path
        self.hierarchy = self.generate_hierarchy()
        self.generated_data = OrderedDict()
        self.interpolation_validator = InterpolationValidator()
        self.type_strategies = type_strategies
        self.fallback_strategies = fallback_strategies
        self.type_conflict_strategies = type_conflict_strategies
        if multi_line_string is True:
            yaml.representer.BaseRepresenter.represent_scalar = ConfigGenerator.custom_represent_scalar

    @staticmethod
    def yaml_dumper():
        try:
            from yaml import CLoader as Loader, CDumper as Dumper
        except ImportError:
            from yaml import Loader, Dumper
        from yaml.representer import SafeRepresenter
        _mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG

        def dict_representer(dumper, data):
            return dumper.represent_dict(iteritems(data))

        def dict_constructor(loader, node):
            return OrderedDict(loader.construct_pairs(node))

        Dumper.add_representer(OrderedDict, dict_representer)
        Loader.add_constructor(_mapping_tag, dict_constructor)

        def str_representer_pipestyle(dumper, data):
            style = '|' if '\n' in data else None
            return dumper.represent_scalar('tag:yaml.org,2002:str', data, style=style)

        Dumper.add_representer(str, str_representer_pipestyle)

        if not PY3:
            def unicode_representer_pipestyle(dumper, data):
                style = u'|' if u'\n' in data else None
                return dumper.represent_scalar(u'tag:yaml.org,2002:str', data, style=style)

            Dumper.add_representer(unicode, unicode_representer_pipestyle)

        return Dumper

    @staticmethod
    def get_yaml_from_path(working_directory, path):
        yaml_files = []
        for yaml_file in os.listdir(path):
            if yaml_file.endswith(".yaml"):
                yaml_files.append(os.path.join(working_directory, yaml_file))
        return sorted(yaml_files)

    @staticmethod
    def yaml_get_content(yaml_file):
        with open(yaml_file, 'r') as f:
            content = yaml.load(f, Loader=yaml.SafeLoader)
        return content if content else {}

    @staticmethod
    def merge_value(reference, new_value, type_strategies, fallback_strategies, type_conflict_strategies):
        merger = Merger(type_strategies, fallback_strategies, type_conflict_strategies)
        if isinstance(new_value, (list, set, dict)):
            new_reference = merger.merge(reference, new_value)
        else:
            raise TypeError("Cannot handle merge_value of type {}".format(type(new_value)))
        return new_reference

    @staticmethod
    def merge_yamls(values, yaml_content, type_strategies, fallback_strategies, type_conflict_strategies):
        for key, value in iteritems(yaml_content):
            if key in values and type(values[key]) != type(value):
                raise Exception("Failed to merge key '{}', because of mismatch in type: {} vs {}"
                                .format(key, type(values[key]), type(value)))
            if key in values and not isinstance(value, primitive_types):
                values[key] = ConfigGenerator.merge_value(values[key], value, type_strategies, fallback_strategies,
                                                          type_conflict_strategies)
            else:
                values[key] = value

    @staticmethod
    def resolve_simple_interpolations(data, current_yaml_file):

        directory = os.path.dirname(current_yaml_file)
        directory = os.path.join(os.getcwd(), directory)

        looper = DictIterator()
        looper.loop_all_items(data, lambda value: replace_parent_working_directory(value, directory))

    def generate_hierarchy(self):
        """
        the method will go through the hierarchy of directories and create an ordered list of directories to be used
        when merging data at each layer
        :return: returns a list of directories in a priority order (from less specific to more specific)
        """
        hierarchy = []
        full_path = pathlib2.Path(self.path)
        for path in full_path.parts:
            os.chdir(path)
            new_path = os.path.relpath(os.getcwd(), self.cwd)
            hierarchy.append(self.get_yaml_from_path(new_path, os.getcwd()))
        os.chdir(self.cwd)
        return hierarchy

    def process_hierarchy(self):
        merged_values = OrderedDict()
        for yaml_files in self.hierarchy:
            for yaml_file in yaml_files:
                yaml_content = self.yaml_get_content(yaml_file)
                self.merge_yamls(merged_values, yaml_content, self.type_strategies, self.fallback_strategies,
                                 self.type_conflict_strategies)
                self.resolve_simple_interpolations(merged_values, yaml_file)
        self.generated_data = merged_values

    def get_values_from_dir_path(self):
        values = {}
        full_path = pathlib2.Path(self.path)
        for path in full_path.parts[1:]:
            split_value = path.split('=')
            values[split_value[0]] = split_value[1]
        return values

    def output_yaml_data(self, data):
        return yaml.dump(data, Dumper=ConfigGenerator.yaml_dumper(), default_flow_style=False, width=200, sort_keys=False)

    def yaml_to_json(self, yaml_data):
        return json.dumps(yaml.load(yaml_data, Loader=yaml.SafeLoader), indent=4)

    def output_data(self, data, output_format):
        yaml_data = self.output_yaml_data(data)
        if "yaml" in output_format:
            return yaml_data
        elif "json" in output_format:
            return self.yaml_to_json(yaml_data)
        raise Exception("Unknown output format: {}".format(output_format))

    def add_enclosing_key(self, key):
        return {key: self.generated_data}

    def remove_enclosing_key(self, key):
        return self.generated_data[key]

    def filter_data(self, keys):
        self.generated_data = {key: self.generated_data[key] for key in keys if key in self.generated_data}

    def exclude_keys(self, keys):
        for key in keys:
            if key in self.generated_data:
                try:
                    logger.info("Excluding key %s", key)
                    del self.generated_data[key]
                except KeyNotFound:
                    logger.info("Excluded key %s not found or already removed", key)

    def add_dynamic_data(self):
        remote_state_retriever = S3TerraformRemoteStateRetriever()
        if "remote_states" in self.generated_data:
            state_files = self.generated_data["remote_states"]
            remote_states = remote_state_retriever.get_dynamic_data(state_files)
            self.merge_value(self.generated_data, remote_states, self.type_strategies, self.fallback_strategies,
                             self.type_conflict_strategies)

    def resolve_interpolations(self):
        resolver = InterpolationResolver()
        self.generated_data = resolver.resolve_interpolations(self.generated_data)

    def resolve_secrets(self, default_aws_profile):
        resolver = SecretResolver()
        self.generated_data = resolver.resolve_secrets(self.generated_data, default_aws_profile)

    def resolve_env(self):
        resolver = EnvVarResolver()
        self.generated_data = resolver.resolve_env_vars(self.generated_data)

    def validate_interpolations(self):
        self.interpolation_validator.check_all_interpolations_resolved(self.generated_data)

    def clean_escape_characters(self):
        """
        Method should clean the escaping characters {{` and `}} from all escaped values
        """
        resolver = EscapingResolver()
        self.generated_data = resolver.resolve_escaping(self.generated_data)

    @staticmethod
    def should_use_block(value):
        """
        https://stackoverflow.com/questions/8640959/how-can-i-control-what-scalar-form-pyyaml-uses-for-my-data
        """

        for c in u"\u000a\u000d\u001c\u001d\u001e\u0085\u2028\u2029":
            if c in value:
                return True
        return False

    @staticmethod
    def custom_represent_scalar(self, tag, value, style=None):
        if style is None:
            if ConfigGenerator.should_use_block(value):
                style = '|'
            else:
                style = self.default_style

        node = yaml.representer.ScalarNode(tag, value, style=style)
        if self.alias_key is not None:
            self.represented_objects[self.alias_key] = node
        return node



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import logging
import os
from .simplessm import SimpleSSM
from .simples3 import SimpleS3
from .simplevault import SimpleVault


class SecretResolver:
    def supports(self, secret_type):
        return False

    def resolve(self, secret_type, secret_params):
        return None

    def get_param_or_exception(self, key, params):
        if key not in params:
            raise Exception("Could not find required key '{}' in the secret params: {}".format(key, params))
        return params[key]


class SSMSecretResolver(SecretResolver):
    def __init__(self, default_aws_profile=None):
        self.default_aws_profile = default_aws_profile

    def supports(self, secret_type):
        return secret_type == "ssm"

    def resolve(self, secret_type, secret_params):
        aws_profile = secret_params.get("aws_profile", self.default_aws_profile)
        if not aws_profile:
            raise Exception(
                "Could not find the aws_profile in the secret params for SSM secret: {}".format(secret_params))

        path = self.get_param_or_exception("path", secret_params)
        region_name = secret_params.get("region_name", "us-east-1")
        ssm = SimpleSSM(aws_profile, region_name)
        return ssm.get(path)


class S3SecretResolver(SecretResolver):
    def __init__(self, default_aws_profile=None):
        self.default_aws_profile = default_aws_profile

    def supports(self, secret_type):
        return secret_type == "s3"

    def resolve(self, secret_type, secret_params):
        aws_profile = secret_params.get("aws_profile", self.default_aws_profile)
        if not aws_profile:
            raise Exception(
                "Could not find the aws_profile in the secret params for S3 secret: {}".format(secret_params))

        bucket = self.get_param_or_exception("bucket", secret_params)
        path = self.get_param_or_exception("path", secret_params)
        region_name = secret_params.get("region_name", "us-east-1")
        base64Encode = secret_params.get("base64encode", False)
        base64Encode = base64Encode == 'true'
        s3 = SimpleS3(aws_profile, region_name)
        return s3.get(bucket, path, base64Encode)


class VaultSecretResolver(SecretResolver):
    def supports(self, secret_type):
        return secret_type == "vault"

    def resolve(self, secret_type, secret_params):
        vault = SimpleVault

        # Generate a token for a policy
        if "token_policy" in secret_params.keys():
            policy = self.get_param_or_exception("token_policy", secret_params)
            return vault().get_token(policy)

        # Retrieve secret from vault path
        if "path" in secret_params.keys():
            path = self.get_param_or_exception("path", secret_params)
            return vault().get_path(path)

        if "key" in secret_params.keys():
            key_path = os.path.split(self.get_param_or_exception("key", secret_params))
            path = key_path[0]
            key = key_path[1]
            return vault().get_key(path, key)


class AggregatedSecretResolver(SecretResolver):
    def __init__(self, default_aws_profile=None):
        self.secret_resolvers = (SSMSecretResolver(default_aws_profile), S3SecretResolver(default_aws_profile),
                                 VaultSecretResolver())

    def supports(self, secret_type):
        return any([resolver.supports(secret_type) for resolver in self.secret_resolvers])

    def resolve(self, secret_type, secret_params):
        for resolver in self.secret_resolvers:
            if resolver.supports(secret_type):
                return resolver.resolve(secret_type, secret_params)

        raise Exception("Could not resolve secret type '{}' with params {}".format(secret_type, secret_params))



import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open('README.md', encoding="utf-8") as f:
    _readme = f.read()

_install_requires = [
    'deepmerge==1.1.0',
    'lru_cache==0.2.3',
    'backports.functools_lru_cache==1.6.6',
    'pathlib2==2.3.7.post1',
    'pyyaml==6.0.1',
    'boto3==1.28.12',
    'hvac==1.1.1'
]

setup(
    name='himl',
    version="0.15.0",
    description='A hierarchical config using yaml',
    long_description=_readme + '\n\n',
    long_description_content_type='text/markdown',
    url='https://github.com/adobe/himl',
    author='Adobe',
    author_email='noreply@adobe.com',
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*",
    license='Apache2',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Markup :: HTML'
    ],
    packages=['himl'],
    include_package_data=True,
    install_requires=_install_requires,
    entry_points={
        'console_scripts': [
            'himl = himl.main:run',
            'himl-config-merger = himl.config_merger:run'
        ]
    }
)



# Copyright 2019 Adobe. All rights reserved.
# This file is licensed to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License. You may obtain a copy
# of the License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed under
# the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR REPRESENTATIONS
# OF ANY KIND, either express or implied. See the License for the specific language
# governing permissions and limitations under the License.

import boto3
import logging
import os
from botocore.exceptions import ClientError


logger = logging.getLogger(__name__)

class SimpleS3(object):
    def __init__(self, aws_profile, region_name):
        self.aws_profile = aws_profile
        self.region_name = region_name

    def get(self, bucket_name, bucket_key, base64Encode=False):
        try:
            logger.info("Resolving S3 object for bucket %s, key '%s' on profile %s in region %s", 
                bucket_name, bucket_key, self.aws_profile, self.region_name)
            client = self.get_s3_client()
            bucket_object = client.get_object(Bucket=bucket_name, Key=bucket_key)["Body"].read()
            return self.parse_data(bucket_object, base64Encode)
        except ClientError as e:
            raise Exception(
                'Error while trying to read S3 value for bucket_name %s, bucket_key: %s - %s' 
                % (bucket_name, bucket_key, e.response['Error']['Code']))

    def parse_data(self, bucket_object, base64Encode):
        if base64Encode:
            import base64
            encodedBytes = base64.b64encode(bucket_object)
            return str(encodedBytes, "utf-8")
        return bucket_object

    def get_s3_client(self):
        session = boto3.session.Session(profile_name=self.aws_profile)
        return session.client('s3')
